/*
 * nlsop.c 
 *
 * author: frank zirkelbach (frank.zirkelbach@physik.uni-augsburg.de)
 *
 * this program tries helping to understand the amorphous depuration
 * and recrystallization of SiCx while ion implantation at temperatures
 * below 400 degree celsius.
 * hopefully the program will simulate the stabilization of the
 * selforganizing lamella structure in the observed behaviour.
 *
 * refs: 
 *  - J. K. N. Lindner. Habil.Schrift, Universitaet Augsburg.
 *  - Maik Haeberlen. Diplomarbeit, Universitaet Augsburg.
 *
 * Copyright (C) 2004 Frank Zirkelbach
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "nlsop.h"

#include "dfbapi.h"
#include "random.h"

#define MAKE_AMORPH(N) *(N)|=AMORPH
#define MAKE_CRYST(N) *(N)&=~AMORPH

int usage(void)
{
 puts("usage:");
 puts("-h \t\t help");
 puts("-n \t\t no user interaction");
 puts("-Z \t\t cryst -> amorph c diffusion in z direction");
 puts("-i \t\t no cryst to cryst diffusion");
 printf("-x <value> \t # x cells (default %d)\n",_X);
 printf("-y <value> \t # y cells (default %d)\n",_Y);
 printf("-z <value> \t # z cells (default %d)\n",_Z);
 printf("-s <value> \t steps (default %d)\n",STEPS);
 printf("-d <value> \t refresh display (default %d)\n",REFRESH);
 printf("-r <value> \t amorphous influence range (default %d)\n",RANGE);
 printf("-f <value> \t stress induced amorphization influence (default %f)\n",S_D);
 printf("-p <value> \t ballistic amorphization influence (default %f)\n",B_D);
 printf("-F <value> \t carbon induced amorphization influence (default %f)\n",C_D);
 printf("-D <value> \t diffusion rate from cryst to amorph cells (default %f)\n",DR_AC);
 printf("-e <value> \t do diffusion every <value> steps (default %d)\n",DIFF_RATE);
 puts("-g <file> <step> continue simulation from file and step (step > 0)!");
 printf("-W <value> \t write every <value> steps to save file (default %d)\n",RESAVE);
 puts("-C <file> \t convert file to gnuplot format");
 puts("-L <file> \t load from file");
 puts("-S <file> \t save to file");
 puts("-R <file> \t read from random file");
 puts("-P <file> \t specify implantation profile file");
 puts("-N <file> \t specify nuclear energy loss profile file");
 printf("-H <value> \t collisions per ion in simulation window (default %d)\n",CPI);
 printf("-m <value> \t number of ion hits sputtering 3nm (default %d)\n",S_RATE);
 
 return 1;
}

int sputter(d3_lattice *d3_l)
{
 int i,size;
 int offh,offl;

 size=d3_l->max_x*d3_l->max_y;
 offl=0;
 offh=size;

 for(i=0;i<d3_l->max_z-1;i++)
 {
  memcpy(d3_l->status+offl,d3_l->status+offh,size);
  memcpy(d3_l->extra+offl,d3_l->extra+offh,size*sizeof(int));
  offl=offh;
  offh+=size;
 }
 memset(d3_l->status+offl,0,size);
 memset(d3_l->extra+offl,0,size);

 return 1;
}

int process_cell(d3_lattice *d3_l,u32 x,u32 y,u32 z,info *my_info,u32 nel_z)
{
 unsigned char *thiz;
 int *conc;
 int i,j;
 int off;
 double p,q;

 thiz=d3_l->status+x+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y;
 conc=d3_l->extra+x+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y;
 p=my_info->b*nel_z;
 for(i=-(my_info->range);i<=my_info->range;i++)
 {
  for(j=-(my_info->range);j<=my_info->range;j++)
  {
   if(!(i==0 && j==0))
   {
    off=((x+d3_l->max_x+i)%d3_l->max_x)+((y+d3_l->max_y+j)%d3_l->max_x)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y;
    if(*(d3_l->status+off)&AMORPH) p+=my_info->s*(*(d3_l->extra+off))*URAND_MAX/(i*i+j*j);
   } 
  }
 }
 p+=*conc*my_info->c*URAND_MAX;
 if(!(*thiz&AMORPH))
 {
  if(get_rand(URAND_MAX)<=p) MAKE_AMORPH(thiz);
 } else
 {
  /* assume 1-p probability */
  /* also look for neighbours ! */
  q=(URAND_MAX-p)>0?URAND_MAX-p:0;
  j=0;
  j+=(*(d3_l->status+((x+d3_l->max_x+1)%d3_l->max_x)+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y)&AMORPH)?1:0;
  j+=(*(d3_l->status+((x+d3_l->max_x-1)%d3_l->max_x)+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y)&AMORPH)?1:0;
  j+=(*(d3_l->status+x+((y+1+d3_l->max_y)%d3_l->max_y)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y)&AMORPH)?1:0;
  j+=(*(d3_l->status+x+((y-1+d3_l->max_y)%d3_l->max_y)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y)&AMORPH)?1:0;
  j+=(*(d3_l->status+x+y*d3_l->max_x+((z+1+d3_l->max_z)%d3_l->max_z)*d3_l->max_x*d3_l->max_y)&AMORPH)?1:0;
  j+=(*(d3_l->status+x+y*d3_l->max_x+((z-1+d3_l->max_z)%d3_l->max_z)*d3_l->max_x*d3_l->max_y)&AMORPH)?1:0;

  p+=((q/6)*j);
  if(get_rand(URAND_MAX)>p) MAKE_CRYST(thiz);
 }
 
 return 1;
}

int distrib_c(d3_lattice *d3_l,info *my_info,int step,u32 rj_m,u32 *rj_g)
{
 u32 x,y,z;
 int i,j,k,c;
 int offset,off;
 int carry;

 /* put one c ion somewhere in the lattice */
 x=get_rand(d3_l->max_x);
 y=get_rand(d3_l->max_y);
 z=get_rand_reject(d3_l->max_z,rj_m,rj_g);
 *(d3_l->extra+x+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y)+=1;
 (my_info->cc)++;

 if(step%my_info->diff_rate==0)
 {

 for(i=0;i<d3_l->max_x;i++)
 {
  for(j=0;j<d3_l->max_y;j++)
  {
   for(k=0;k<d3_l->max_z;k++)
   {
    offset=i+j*d3_l->max_x+k*d3_l->max_x*d3_l->max_y;
    /* case amorph: amorph <- cryst diffusion */
    if(*(d3_l->status+offset)&AMORPH)
    {
     for(c=-1;c<=1;c++)
     {
      if(c!=0)
      {
       off=((i+d3_l->max_x+c)%d3_l->max_x)+j*d3_l->max_x+k*d3_l->max_x*d3_l->max_y;
       carry=0;
       if(!(*(d3_l->status+off)&AMORPH)) carry=(int)(my_info->dr_ac*(*(d3_l->extra+off)));
       if(carry!=0)
       {
        *(d3_l->extra+offset)+=carry;
        *(d3_l->extra+off)-=carry;
       }
      }
     }
     for(c=-1;c<=1;c++)
     {
      if(c!=0)
      {
       off=i+((j+c+d3_l->max_y)%d3_l->max_y)*d3_l->max_x+k*d3_l->max_x*d3_l->max_y;
       carry=0;
       if(!(*(d3_l->status+off)&AMORPH)) carry=(int)(my_info->dr_ac*(*(d3_l->extra+off)));
       if(carry!=0)
       {
        *(d3_l->extra+offset)+=carry; 
        *(d3_l->extra+off)-=carry; 
       }
      }
     }
     /* z diff */
     if(k!=0)
     {
      off=i+j*d3_l->max_x+(k-1)*d3_l->max_x*d3_l->max_y;
      carry=0;
      if(!*(d3_l->status+off)&AMORPH) carry=(int)(my_info->dr_ac*(*(d3_l->extra+off)));
      if(carry!=0)
      {
       *(d3_l->extra+off)-=carry;
       *(d3_l->extra+offset)+=carry;
      }
     }
     if(k!=d3_l->max_z-1)
     {
      off=i+j*d3_l->max_x+(k+1)*d3_l->max_x*d3_l->max_y;
      carry=0;
      if(!*(d3_l->status+off)&AMORPH) carry=(int)(my_info->dr_ac*(*(d3_l->extra+off)));
     if(carry!=0)
      {
       *(d3_l->extra+off)-=carry;
       *(d3_l->extra+offset)+=carry;
      }
     }
    }
   } /* for z */
  } /* for y */
 } /* for x */

 } /* if step modulo diff_rate == 0 */

 return 1;
}

int calc_pressure(d3_lattice *d3_l,int range)
{
 int i,j,off;
 double count,max=0;
 int x,y,z;

 for(x=0;x<d3_l->max_x;x++)
 {
  for(y=0;y<d3_l->max_y;y++)
  {
   for(z=0;z<d3_l->max_z;z++)
   {
    count=0;
    for(i=-range;i<=range;i++)
    {
     for(j=-range;j<=range;j++)
     {
      if(i!=0 && j!=0)
      {
       off=((x+d3_l->max_x+i)%d3_l->max_x)+((y+d3_l->max_y+j)%d3_l->max_x)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y;
       if(*(d3_l->status+off)&AMORPH) count+=((double)*(d3_l->extra+off))/(i*i+j*j);
      }
     }
    }
    if(count>max) max=count;
   }
  }
 }

 for(x=0;x<d3_l->max_x;x++)
 {
  for(y=0;y<d3_l->max_y;y++)
  {
   for(z=0;z<d3_l->max_z;z++)
   {
    count=0;
    for(i=-range;i<=range;i++)
    {
     for(j=-range;j<=range;j++)
     {
      if(i!=0 && j!=0)
      {
       off=((x+d3_l->max_x+i)%d3_l->max_x)+((y+d3_l->max_y+j)%d3_l->max_x)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y;
       if(*(d3_l->status+off)&AMORPH) count+=((double)*(d3_l->extra+off))/(i*i+j*j);
      }
     }
    }
    *((unsigned int *)(d3_l->v_ptr)+x+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y)=(count*500/max);
   }
  }
 }

 return 1;
}

int calc_max_extra(d3_lattice *d3_l)
{
 int x,y,z;
 int off,max=0;

 for(x=0;x<d3_l->max_x;x++)
 {
  for(y=0;y<d3_l->max_y;y++)
  {
   for(z=0;z<d3_l->max_z;z++)
   {
    off=x+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y;
    if(*(d3_l->extra+off)>max) max=*(d3_l->extra+off);
   }
  }
 }

 return max;
}

int write_ac_distr(d3_lattice *d3_l,int ac_distr)
{
 int fd,x,y,z;
 int count,offset;
 char file[32];
 int si_count;
 
 if(ac_distr==1) strcpy(file,"carbon_in_av.plot");
 if(ac_distr==2) strcpy(file,"carbon_in_cv.plot");
 if(ac_distr==3) strcpy(file,"carbon.plot");
 if(ac_distr==4) strcpy(file,"amorphous_volumes.plot");

 if((fd=open(file,O_WRONLY|O_CREAT))<0)
 {
  puts("cannot open plot file");
  return -1;
 }

 for(z=0;z<d3_l->max_z;z++)
 {
  count=0;
  si_count=0;
  for(x=0;x<d3_l->max_x;x++)
  {
   for(y=0;y<d3_l->max_y;y++)
   {
    offset=x+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y;
    if(ac_distr==1)
     if(*(d3_l->status+offset)&AMORPH)
     {
      count+=*(d3_l->extra+offset);
      si_count+=1;
     }
    if(ac_distr==2)
     if(!(*(d3_l->status+offset)&AMORPH))
     {
      count+=*(d3_l->extra+offset);
      si_count+=1;
     }
    if(ac_distr==3)
    {
     count+=*(d3_l->extra+offset);
     si_count+=1;
    }
    if(ac_distr==4)
     if(*(d3_l->status+offset)&AMORPH) count+=1;
   }
  }
  si_count*=SI_PER_VOLUME;
  if(ac_distr==4) dprintf(fd,"%d %d\n",z*CELL_LENGTH,count);
#ifdef ATPROZ
  else
  {
   if(si_count)
    dprintf(fd,"%d %f\n",z*CELL_LENGTH,100.0*count/(si_count+count));
   else
    dprintf(fd,"%d 0\n",z*CELL_LENGTH);
  }
#else
  else dprintf(fd,"%d %d\n",z*CELL_LENGTH,count/(d3_l->max_x*d3_l->max_y));
#endif
 }
 close(fd);
 
 return 1;
}

int write_bmp(d3_lattice *d3_l,int window,u32 x,u32 y,u32 z,int max)
{
 int fd,i,j,size=0,foo=0,k,sum;
 int width=0,height=0;
 char bmpfile[MAX_CHARS];
 char buf[128];

 if(window==1)
 {
  sprintf(bmpfile,"x-z_%d.bmp",y);
  foo=3*d3_l->max_x;
  size=(foo+(4-foo%4))*d3_l->max_z;
  height=d3_l->max_z;
  width=d3_l->max_x;
 }
 if(window==2)
 {
  sprintf(bmpfile,"y-z_%d.bmp",x);
  foo=3*d3_l->max_y;
  size=(foo+(4-foo%4))*d3_l->max_z;
  height=d3_l->max_z;
  width=d3_l->max_y;
 }
 if(window==3)
 {
  sprintf(bmpfile,"x-y_%d.bmp",z);
  foo=3*d3_l->max_x;
  size=(foo+(4-foo%4))*d3_l->max_y;
  width=d3_l->max_x;
  height=d3_l->max_y;
 }
 if(window==4)
 {
  sprintf(bmpfile,"x-z_cdistr_%d.bmp",y);
  foo=3*d3_l->max_x;
  size=(foo+(4-foo%4))*d3_l->max_z;
  height=d3_l->max_z;
  width=d3_l->max_x;
 }
 if(window==5)
 {
  sprintf(bmpfile,"y-z_cdistr_%d.bmp",x);
  foo=3*d3_l->max_y;
  size=(foo+(4-foo%4))*d3_l->max_z;
  height=d3_l->max_z;
  width=d3_l->max_y;
 }
 if(window==6)
 {
  sprintf(bmpfile,"x-y_cdistr_%d.bmp",z);
  foo=3*d3_l->max_x;
  size=(foo+(4-foo%4))*d3_l->max_y;
  width=d3_l->max_x;
  height=d3_l->max_y;
 }
 if(window==7)
 {
  sprintf(bmpfile,"x-z_pressure_%d.bmp",y);
  foo=3*d3_l->max_x;
  size=(foo+(4-foo%4))*d3_l->max_z;
  height=d3_l->max_z;
  width=d3_l->max_x;
 }
 if(window==8)
 {
  sprintf(bmpfile,"y-z_pressure_%d.bmp",x);
  foo=3*d3_l->max_y;
  size=(foo+(4-foo%4))*d3_l->max_z;
  height=d3_l->max_z;
  width=d3_l->max_y;
 }
 if(window==9)
 {
  sprintf(bmpfile,"x-y_pressure_%d.bmp",z);
  foo=3*d3_l->max_x;
  size=(foo+(4-foo%4))*d3_l->max_y;
  height=d3_l->max_y;
  width=d3_l->max_x;
 }
 if(window==10)
 {
  sprintf(bmpfile,"x-z_c_cdistr_%d.bmp",y);
  foo=3*d3_l->max_x;
  size=(foo+(4-foo%4))*d3_l->max_z;
  height=d3_l->max_z;
  width=d3_l->max_x;
 }
 if(window==11)
 {
  sprintf(bmpfile,"y-z_c_cdistr_%d.bmp",x);
  foo=3*d3_l->max_y;
  size=(foo+(4-foo%4))*d3_l->max_z;
  height=d3_l->max_z;
  width=d3_l->max_y;
 }
 if(window==12)
 {
  sprintf(bmpfile,"x-y_c_cdistr_%d.bmp",z);
  foo=3*d3_l->max_x;
  size=(foo+(4-foo%4))*d3_l->max_y;
  height=d3_l->max_y;
  width=d3_l->max_x;
 }
 if((fd=open(bmpfile,O_WRONLY|O_CREAT))<0)
 {
  puts("cannot open bmp file");
  return -1;
 }

 /* bmpheader */
 buf[0]='B'; /* std header start */
 buf[1]='M';
 buf[2]=(size+0x36)&0xff; /* file size */
 buf[3]=(size+0x36)>>8;
 memset(buf+4,0,6);
 buf[10]=0x36; /* offset to data */
 memset(buf+11,0,3);
 buf[14]=0x28; /* length of bmp info header */
 memset(buf+15,0,3);
 buf[18]=width&0xff; /* width and height */
 buf[19]=width>>8;
 memset(buf+20,0,2);
 buf[22]=height&0xff;
 buf[23]=height>>8;
 memset(buf+24,0,2);
 buf[26]=1; /* # planes -> 1 */
 buf[27]=0;
 buf[28]=24; /* bits per pixel -> 2^24 (true color) */
 buf[29]=0;
 memset(buf+30,0,4); /* compression -> none */
 buf[34]=size&0xff; /* data size */
 buf[35]=size>>8;
 memset(buf+36,0,2);
 buf[38]=0x12; /* res: pixel/meter */
 buf[39]=0x0b;
 memset(buf+40,0,2);
 buf[42]=0x12;
 buf[43]=0x0b;
 memset(buf+44,0,2); 
 memset(buf+46,0,8); /* no colors, no important colors */

 if(write(fd,buf,54)<54)
 {
  puts("failed writing bmp header");
  return -1;
 }
 if(window==1)
 {
  for(j=0;j<d3_l->max_z;j++)
  {
   for(i=0;i<d3_l->max_x;i++)
   {
    sum=0;
    for(k=-2;k<=2;k++)
     if(*(d3_l->status+i+(((y+k+d3_l->max_y)%d3_l->max_y)*d3_l->max_x)+(d3_l->max_z-j-1)*d3_l->max_x*d3_l->max_y)&RED) sum+=0xff;
    sum/=5;
    memset(buf,(unsigned char)sum,3);
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return-1;
    }
   }
   if(foo%4)
   {
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   } 
  }
 }
 if(window==2)
 {
  for(j=0;j<d3_l->max_z;j++)
  {
   for(i=0;i<d3_l->max_y;i++)
   {
    sum=0;
    for(k=-2;k<=2;k++)
     if(*(d3_l->status+((x+k+d3_l->max_x)%d3_l->max_x)+i*d3_l->max_x+(d3_l->max_z-j-1)*d3_l->max_x*d3_l->max_y)&RED) sum+=0xff;
    sum/=5;
    memset(buf,(unsigned char)sum,3);
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return-1;
    }
   }
   if(foo%4)
   {
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   }
  }
 }
 if(window==3)
 {
  for(j=0;j<d3_l->max_y;j++)
  {
   for(i=0;i<d3_l->max_x;i++)
   {
    if(*(d3_l->status+i+(d3_l->max_y-j-1)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y)&RED) memset(buf,0xff,3);
    else memset(buf,0,3);
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return -1;
    }
   }
   if(foo%4)
   {
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   }
  }
 }
 if(window==4)
 {
  for(j=0;j<d3_l->max_z;j++)
  {
   for(i=0;i<d3_l->max_x;i++)
   {
    sum=*(d3_l->extra+i+y*d3_l->max_x+(d3_l->max_z-j-1)*d3_l->max_x*d3_l->max_y);
    sum=sum*500/max;
    buf[2]=0;
    if(sum<=255) {
      buf[1]=0;
      buf[0]=sum;
    } else {
      buf[1]=(sum-255);
      buf[0]=0xff;
    }
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return-1;
    }
   }
   if(foo%4)
   {
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   }
  }
 }
 if(window==5)
 { 
  for(j=0;j<d3_l->max_z;j++) 
  {                     
   for(i=0;i<d3_l->max_x;i++)
   {
    sum=*(d3_l->extra+x+i*d3_l->max_x+(d3_l->max_z-j-1)*d3_l->max_x*d3_l->max_y);
    sum=sum*500/max;
    buf[2]=0;
    if(sum<=255) {
      buf[1]=0;
      buf[0]=sum;
    } else {
      buf[1]=(sum-255);
      buf[0]=0xff;
    }
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return-1;  
    }
   }
   if(foo%4)
   {
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   }
  }
 }
 if(window==6)
 {   
  for(j=0;j<d3_l->max_y;j++)
  {  
   for(i=0;i<d3_l->max_x;i++)
   { 
    sum=*(d3_l->extra+i+(d3_l->max_y-j-1)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y);
    sum=sum*500/max;
    buf[2]=0;
    if(sum<=255) {
      buf[1]=0;
      buf[0]=sum;
    } else {
      buf[1]=(sum-255);
      buf[0]=0xff;
    }
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return-1;
    }
   } 
   if(foo%4)
   { 
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   } 
  }
 }
 if(window==7)
 {
  for(j=0;j<d3_l->max_z;j++)
  {
   for(i=0;i<d3_l->max_x;i++)
   {
    sum=*((unsigned int *)(d3_l->v_ptr)+i+y*d3_l->max_x+(d3_l->max_z-j-1)*d3_l->max_x*d3_l->max_y);
      buf[0]=0;
    if(sum<=255) {
      buf[1]=0;
      buf[2]=sum;
    } else {
      buf[1]=(sum-255);
      buf[2]=0xff;
    }
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return-1;
    }
   }
   if(foo%4)
   {
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   }
  }
 }
 if(window==8)
 { 
  for(j=0;j<d3_l->max_z;j++) 
  {                     
   for(i=0;i<d3_l->max_x;i++)
   {
    sum=*((unsigned int *)(d3_l->v_ptr)+x+i*d3_l->max_x+(d3_l->max_z-j-1)*d3_l->max_x*d3_l->max_y);
      buf[0]=0;
    if(sum<=255) {
      buf[1]=0;
      buf[2]=sum;
    } else {
      buf[1]=(sum-255);
      buf[2]=0xff;
    }
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return-1;  
    }
   }
   if(foo%4)
   {
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   }
  }
 }
 if(window==9)
 {   
  for(j=0;j<d3_l->max_y;j++)
  {  
   for(i=0;i<d3_l->max_x;i++)
   {
    sum=*((unsigned int *)(d3_l->v_ptr)+i+(d3_l->max_y-j-1)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y);
      buf[0]=0;
    if(sum<=255) {
      buf[1]=0;
      buf[2]=sum;
    } else {
      buf[1]=(sum-255);
      buf[2]=0xff;
    }
    printf("sum = %d => r: %02x g: %02x b: %02x\n",sum,buf[2],buf[1],buf[0]);
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return-1;
    }
   } 
   if(foo%4)
   { 
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   } 
  }
 }
 if(window==10)
 {
  for(j=0;j<d3_l->max_z;j++)
  {
   for(i=0;i<d3_l->max_x;i++)
   {
    if(!(*(d3_l->status+i+y*d3_l->max_x+(d3_l->max_z-j-1)*d3_l->max_x*d3_l->max_y)&RED)) sum=*(d3_l->extra+i+y*d3_l->max_x+(d3_l->max_z-j-1)*d3_l->max_x*d3_l->max_y);
    else sum=0;
    sum=sum*255/max;
    memset(buf,(unsigned char)sum,3);
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return-1;
    }
   }
   if(foo%4)
   {
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   }
  }
 }
 if(window==11)
 { 
  for(j=0;j<d3_l->max_z;j++) 
  {                     
   for(i=0;i<d3_l->max_x;i++)
   {
    if(!(*(d3_l->status+x+i*d3_l->max_x+(d3_l->max_z-j-1)*d3_l->max_x*d3_l->max_y)&RED)) sum=*(d3_l->extra+x+i*d3_l->max_x+(d3_l->max_z-j-1)*d3_l->max_x*d3_l->max_y);
    else sum=0;
    sum=sum*255/max;
    memset(buf,(unsigned char)sum,3);
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return-1;  
    }
   }
   if(foo%4)
   {
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   }
  }
 }
 if(window==12)
 {   
  for(j=0;j<d3_l->max_y;j++)
  {  
   for(i=0;i<d3_l->max_x;i++)
   {
    if(!(*(d3_l->status+i+(d3_l->max_y-j-1)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y)&RED)) sum=*(d3_l->extra+i+(d3_l->max_y-j-1)*d3_l->max_x+z*d3_l->max_x*d3_l->max_y);
    else sum=0;
    sum=sum*255/max;
    memset(buf,(unsigned char)sum,3);
    if(write(fd,buf,3)<3)
    {
     puts("failed writing rgb values to bmp file");
     return-1;
    }
   } 
   if(foo%4)
   { 
    memset(buf,0,4-foo%4);
    if(write(fd,buf,4-foo%4)<4-foo%4)
    {
     puts("failed writing 4 byte ending");
     return -1;
    }
   }
  }
 }
 close(fd);

 return 1;
}

int save_to_file(char *sf,d3_lattice *d3_l,info *my_inf)
{
 int sf_fd,c;

 if((sf_fd=open(sf,O_WRONLY|O_CREAT))<0)
 {
  puts("cannot open save file");
  return -1;
 }
 if(write(sf_fd,d3_l,sizeof(d3_lattice))<sizeof(d3_lattice))
 {
  puts("failed saving d3 lattice struct");
  return -1;
 }
 if(write(sf_fd,my_inf,sizeof(info))<sizeof(info))
 {
  puts("failed saving info struct");
  return-1;
 }
 c=d3_l->max_x*d3_l->max_y*d3_l->max_z;
 if(write(sf_fd,d3_l->status,c*sizeof(unsigned char))<c*sizeof(unsigned char))
 {
  puts("failed saving status of d3 lattice sites");
  return -1;
 }
 if(write(sf_fd,d3_l->extra,c*sizeof(int))<c*sizeof(int))
 {
  puts("failed saving sites concentration");
  return -1;
 }
 close(sf_fd);

 return 1;
}

int load_from_file(char *lf,d3_lattice *d3_l,info *my_inf)
{

 int lf_fd,c,pos,end,data,data_len,strip;

 if((lf_fd=open(lf,O_RDONLY))<0)
 {
  puts("cannot open load file");
  return -1;
 }
 if(read(lf_fd,d3_l,sizeof(d3_lattice))<sizeof(d3_lattice))
 {
  puts("failed reading d3 lattice struct");
  return -1;
 }
 pos=lseek(lf_fd,0,SEEK_CUR);
 printf("psition: %d (%d)\n",pos,sizeof(d3_lattice));
 data=d3_l->max_x*d3_l->max_y*d3_l->max_z;
 data_len=data*(sizeof(int)+sizeof(unsigned char));
 printf("there are %d volumes so we need %d of bytes\n",data,data_len);
 end=lseek(lf_fd,0,SEEK_END);
 c=end-pos-data_len;
 printf("end: %d => length: %d => guessed info size: %d bytes\n",end,end-pos,c);
 strip=sizeof(info)-c;
 printf("as real programs info size is %d, we strip %d bytes\n",sizeof(info),strip);
 lseek(lf_fd,pos,SEEK_SET);
 c=sizeof(info);
 if(strip>0) c-=strip;
 if(c<0)
 {
  puts("info smaller then strip size");
  return -1;
 }
 if(read(lf_fd,my_inf,c)<c)
 {
  puts("failed reading info struct");
  return-1;
 }
 if(strip>0) memset(my_inf+c,0,strip);
 if(strip<0) lseek(lf_fd,(-1*strip),SEEK_CUR);
 c=d3_l->max_x*d3_l->max_y*d3_l->max_z;
 if((d3_l->status=(unsigned char*)malloc(c*sizeof(unsigned char)))==NULL)
 {
  puts("cannot allocate status buffer");
  return -1;
 }
 if((d3_l->extra=(int *)malloc(c*sizeof(int)))==NULL)
 {
  puts("cannot allocate concentration buffer");
  return -1;
 }
 if(read(lf_fd,d3_l->status,c*sizeof(unsigned char))<c*sizeof(unsigned char))
 {
  puts("failed reading status of d3 lattice sites");
  return -1;
 }
 if(read(lf_fd,d3_l->extra,c*sizeof(int))<c*sizeof(int))
 {
  puts("failed reading sites concentration");
  return -1;
 }
 close(lf_fd);

 return 1;
}

int convert_file(char *cf,d3_lattice *d3_l)
{
 int x,y,z;
 int c_fd;

 if((c_fd=open(cf,O_WRONLY|O_CREAT))<0)
 {
  puts("cannot open convert file");
  return -1;
 }
 dprintf(c_fd,"# created by nlsop (gnuplot format)\n");
 for(x=0;x<d3_l->max_x;x++)
 {
  for(y=0;y<d3_l->max_y;y++)
  {
   for(z=0;z<d3_l->max_z;z++)
   {
    if(*(d3_l->status+x+y*d3_l->max_x+z*d3_l->max_x*d3_l->max_y)&AMORPH) dprintf(c_fd,"%d %d %d\n",x,y,z);
   }
  }
 }
 close(c_fd);

 return 1;
}

int get_c_ratio(double *c_ratio,char *pfile,info *my_info,d3_lattice *d3_l)
{
 double all,a,b,d;
 int i,k;
 int p_fd;
 unsigned char buf[32];
 char *p;
 unsigned char c;

 if((p_fd=open(pfile,O_RDONLY))<0)
 {
  puts("cannot open profile file");
  return -1;
 }
 k=1;
 d=0;
 all=0;
 while(k)
 {
  for(i=0;i<32;i++)
  {
   k=read(p_fd,&c,1);
   buf[i]=c;
   if(c=='\n') break;
  }
  if(k)
  {
   p=strtok(buf," ");
   a=atof(p)/10; /* nm */
   p=strtok(NULL," ");
   b=atof(p);
   if(a<=d3_l->max_z*CELL_LENGTH) d+=b;
   all+=b;
  }
 }
 *c_ratio=d/all;
 close(p_fd);

 return 1;
}

u32 get_reject_graph(info *my_info,d3_lattice *d3_l,char *file,u32 *graph) {
 double a,b;
 int i,j,k;
 int fd;
 char buf[32],*p;
 unsigned char *flag;
 u32 max;

 max=0;
 if((fd=open(file,O_RDONLY))<0)
 {
  puts("cannot open file to calculate rejection graph");
  return -1;
 }
 if((flag=(unsigned char *)malloc(d3_l->max_z))==NULL)
 {
  puts("cannot malloc flag memory for rejection graph");
  return -1;
 }
 memset(flag,0,d3_l->max_z);
 memset(graph,0,d3_l->max_z*sizeof(u32));
 /* get fixpoints */
 k=1;
 while(k)
 {
  for(i=0;i<32;i++)
  {
   k=read(fd,&buf[i],1);
   if((buf[i]=='\n')||(k==0)) break;
  }
  if(k)
  {
   p=strtok(buf," ");
   a=atof(p)/10; /* nm */
   p=strtok(NULL," ");
   b=atof(p);
   if(a>d3_l->max_z*CELL_LENGTH) k=0;
   else 
   {
    graph[(int)(a/CELL_LENGTH)]=(int)(URAND_MAX/100*b);
    flag[(int)(a/CELL_LENGTH)]=1;
   }
  }
 }
 /* do (linear) interpolation here! */
 i=0;
 a=0;
 while(i<d3_l->max_z)
 {
  /* graph[0] is 0! */
  j=i;
  i++;
  while(flag[i]==0&&i<d3_l->max_z) i++;
  for(k=j+1;k<i;k++) graph[k]=(int)((k-j)*((int)graph[i]-(int)graph[j])/(i-j))+graph[j];
  if(graph[i]>max) max=graph[i];
 }

 free(flag);

#ifdef DEBUG_INTERPOL_PROFILE
 printf("debug: %s (interpolated profile)\n",file);
 for(i=0;i<d3_l->max_z;i++) printf("%d %d\n",i,graph[i]);
#endif

 return max;
}

int get_amorphous_layer_info(d3_lattice *d3_l,int *sai,int *sacl,int *eacl) {
  int i,j,a,oend,nend,count;
  unsigned char sacl_is_set=0;
  unsigned char eacl_is_set=0;
  unsigned char sai_is_set=0;

  a=d3_l->max_x*d3_l->max_y;
  nend=a;
  oend=0;

  for(i=0;i<d3_l->max_z;i++) {
    count=0;
    for(j=oend;j<nend;j++) if(*(d3_l->status+j)&AMORPH) count++;
    oend=nend;
    nend+=a;
    if((count>=A_START*a)&&(!sai_is_set)) {
      *sai=i;
      sai_is_set=1;
    }
    if((count>=AC_START*a)&&(!sacl_is_set)) {
      *sacl=i;
      sacl_is_set=1;
    }
    if((count<=A_END*a)&&(sacl_is_set)&&(!eacl_is_set)) {
      *eacl=i;
      eacl_is_set=1;
    }
    if((eacl_is_set)&&(count>=A_END*a)) eacl_is_set=0;
  }
  return 1;
}

int main(int argc,char **argv)
{
 u32 x,y,z,x_c,y_c,z_c;
 int i,j,quit,escape,switchmode,nowait,bmp,ac_distr;
 int refresh,resave;
 int c_step;
 int sai,sacl,eacl;
 char s_file[MAX_CHARS];
 char s_file_tmp[MAX_CHARS];
 char l_file[MAX_CHARS];
 char c_file[MAX_CHARS];
 char p_file[MAX_CHARS];
 char n_e_file[MAX_CHARS];
 char convert;
 char r_file[MAX_CHARS];
#ifdef USE_DFB_API
 char xyz_txt[MAX_TXT];
 char status_txt[MAX_TXT];
 char conc_txt[MAX_TXT];
 char steps_txt[MAX_TXT];
 char cc_txt[MAX_TXT];
 char a_txt[MAX_TXT];
 char s_txt[MAX_TXT];
 char ballistic_txt[MAX_TXT];
 char carbon_txt[MAX_TXT];
 char stress_txt[MAX_TXT];
 char r_txt[MAX_TXT];
 char zdiff_txt[MAX_TXT];
 char diff_txt[MAX_TXT];
 char dr_ac_txt[MAX_TXT];
 char dose_txt[MAX_TXT];
 char mode_txt[MAX_TXT];
 char hpi_txt[MAX_TXT];
 char srate_txt[MAX_TXT];
 char start_ai_txt[MAX_TXT];
 char start_acl_txt[MAX_TXT];
 char end_acl_txt[MAX_TXT];
 char *arg_v[MAX_ARGV];
#endif
 d3_lattice d3_l;
 info my_info;
 unsigned char mode;
 double c_ratio;
#ifdef USE_DFB_API
 int max_extra;
#endif
 u32 *c_profile;
 u32 *n_e_loss;
 u32 ne_max,ip_max;
 u32 nel_z;
 unsigned char do_sputter;

 d3_l.max_x=_X;
 d3_l.max_y=_Y;
 d3_l.max_z=_Z;
 my_info.steps=STEPS;
 my_info.range=RANGE;
 refresh=REFRESH;
 resave=RESAVE;
 my_info.s=S_D;
 my_info.b=B_D;
 my_info.c=C_D;
 my_info.cc=CC;
 my_info.dr_ac=DR_AC;
 my_info.diff_rate=DIFF_RATE;
 my_info.cpi=CPI;
 my_info.s_rate=S_RATE;
 nowait=0;
 quit=0;
 escape=0;
 switchmode=0;
 c_step=0;
 strcpy(s_file,"");
 strcpy(l_file,"");
 strcpy(c_file,"");
 strcpy(p_file,IMP_PROFILE);
 strcpy(n_e_file,NEL_PROFILE);
 convert=0;
 strcpy(r_file,"");
 mode=0;
 ne_max=0;
 ip_max=0;
 sai=0;
 sacl=0;
 eacl=0;
 do_sputter=1;

#ifdef MORE_PRINTF
 printf("reading argv ...");
#endif

 for(i=1;i<argc;i++)
 {
  if(argv[i][0]=='-')
  {
   switch(argv[i][1])
   {
    case 'h':
     usage();
     return -1;
    case 'n':
     nowait=1;
     break;
    case 'x':
     d3_l.max_x=atoi(argv[++i]);
     break;
    case 'y':
     d3_l.max_y=atoi(argv[++i]);
     break;
    case 'z':
     d3_l.max_z=atoi(argv[++i]);
     break;
    case 's':
     my_info.steps=atoi(argv[++i]);
     break;
    case 'd':
     refresh=atoi(argv[++i]);
     break;
    case 'r':
     my_info.range=atoi(argv[++i]);
     break;
    case 'f':
     my_info.s=atof(argv[++i]);
     break;
    case 'p':
     my_info.b=atof(argv[++i]);
     break;
    case 'F':
     my_info.c=atof(argv[++i]);
     break;
    case 'W':
     resave=atoi(argv[++i]);
     break;
    case 'C':
     strcpy(l_file,argv[++i]);
     if(i<argc-1) if(argv[i+1][0]!='-') strcpy(c_file,argv[++i]);
     convert=1;
     break;
    case 'D':
     my_info.dr_ac=atof(argv[++i]);
     break;
    case 'e':
     my_info.diff_rate=atoi(argv[++i]);
     break;
    case 'L':
     strcpy(l_file,argv[++i]);
     break;
    case 'S':
     strcpy(s_file,argv[++i]);
     break;
    case 'R':
     strcpy(r_file,argv[++i]);
     break;
    case 'P':
     strcpy(p_file,argv[++i]);
     break;
    case 'N':
     strcpy(n_e_file,argv[++i]);
     break;
    case 'g':
     strcpy(l_file,argv[++i]);
     if(i<argc-1) if(argv[i+1][0]!='-') c_step=atoi(argv[++i]);
     break;
    case 'H':
     my_info.cpi=atoi(argv[++i]);
     break;
    case 'm':
     my_info.s_rate=atoi(argv[++i]);
     break;
    default:
     usage();
     return -1;
   }
  } else usage();
 }

#ifdef MORE_PRINTF
 printf(" done\n");
#endif

 x=d3_l.max_x/2-1;
 y=d3_l.max_y/2-1;
 z=d3_l.max_z/2-1;

#ifdef NODFB
 if(!strcmp(s_file,""))
 {
  puts("NODFB defined, run with -S option");
  return -1;
 }
#endif

 printf("random api init ...");
 if(!strcmp(r_file,"")) rand_init(NULL);
 else rand_init(r_file);
 printf(" done\n");

 if(!strcmp(l_file,""))
 {
  i=d3_l.max_x*d3_l.max_y*d3_l.max_z;
#ifdef USE_DFB_API
  printf("d3 lattice init ...");
  d3_lattice_init(&argc,argv,&d3_l);
  printf(" done\n");
#endif
  if((d3_l.status=(unsigned char *)malloc(i*sizeof(unsigned char)))==NULL)
  {
   puts("failed allocating status buffer");
   return -1;
  }
  memset(d3_l.status,0,i*sizeof(unsigned char));
  if((d3_l.extra=(int *)malloc(i*sizeof(int)))==NULL)
  {
   puts("failed allocating concentration buffer");
   return -1;
  }
  memset(d3_l.extra,0,i*sizeof(int));
 } else
 {
  printf("loading file ...");
  load_from_file(l_file,&d3_l,&my_info);
  printf(" done\n");
  if(convert) 
  {   
   if(!strcmp(c_file,"")) sprintf(c_file,"%s_gnuplot",l_file);
   printf("converting file %s to %s\n",l_file,c_file);
   convert_file(c_file,&d3_l);
   puts("done");
   return 1;
  }
#ifdef USE_DFB_API
  printf("d3 lattice init ...");
  d3_lattice_init(&argc,argv,&d3_l);
  printf(" done\n");
#endif
 }

#ifdef USE_DFB_API
 printf("event init ...");
 d3_event_init(&d3_l);
 printf(" done\n");
#endif

#ifdef USE_DFB_API
 strcpy(a_txt,"args:");
 sprintf(s_txt,"steps: %d",my_info.steps);
 sprintf(dose_txt,"dose: %.2fe+17 C/cm",my_info.steps*1.0/(d3_l.max_x*d3_l.max_y*CELL_LENGTH*CELL_LENGTH*1000));
 sprintf(r_txt,"stress influence range: %d",my_info.range);
 sprintf(stress_txt,"stress term: %f",my_info.s);
 sprintf(ballistic_txt,"ballistic term: %f",my_info.b);
 sprintf(carbon_txt,"carbon term: %f",my_info.c);
 sprintf(dr_ac_txt,"a/c diffusion rate: %f",my_info.dr_ac);
 sprintf(zdiff_txt,"diffusion in z direction: yes");
 sprintf(diff_txt,"diffusion every %d steps",my_info.diff_rate);
 strcpy(mode_txt,"view: a/c mode");
 sprintf(hpi_txt,"hits per ion: %d",my_info.cpi);
 sprintf(srate_txt,"sputter rate (3nm/#c): %d",my_info.s_rate);
 get_amorphous_layer_info(&d3_l,&sai,&sacl,&eacl);
 sprintf(start_ai_txt,"start of a-inclusions: %d nm (%d)",3*sai,sai);
 sprintf(start_acl_txt,"start of a-layer: %d nm (%d)",3*sacl,sacl);
 sprintf(end_acl_txt,"end of a-layer: %d nm (%d)",3*eacl,eacl);
 arg_v[1]=mode_txt;
 arg_v[2]=xyz_txt;
 arg_v[3]=status_txt;
 arg_v[4]=conc_txt;
 arg_v[5]=NULL;
 arg_v[6]=steps_txt;
 arg_v[7]=cc_txt;
 arg_v[8]=NULL;
 arg_v[9]=a_txt;
 arg_v[10]=s_txt;
 arg_v[11]=dose_txt;
 arg_v[12]=diff_txt;
 arg_v[13]=zdiff_txt;
 arg_v[14]=dr_ac_txt;
 arg_v[15]=ballistic_txt;
 arg_v[16]=carbon_txt;
 arg_v[17]=stress_txt;
 arg_v[18]=r_txt;
 arg_v[19]=hpi_txt;
 arg_v[20]=srate_txt;
 arg_v[21]=NULL;
 arg_v[22]=start_ai_txt;
 arg_v[23]=start_acl_txt;
 arg_v[24]=end_acl_txt;
 arg_v[25]=NULL;
#endif

 /* compute graphs for random number rejection method */
 printf("random rejection graphs ...");
 if((c_profile=(u32 *)malloc(d3_l.max_z*sizeof(unsigned int)))==NULL)
 {
  puts("failed allocating memory for carbon profile graph");
  return -1;
 }
 if((n_e_loss=(u32 *)malloc(d3_l.max_z*sizeof(unsigned int)))==NULL)
 {
  puts("failed allocating memory for nuclear energy loss graph");
  return -1;
 }
 ip_max=get_reject_graph(&my_info,&d3_l,p_file,c_profile);
 ne_max=get_reject_graph(&my_info,&d3_l,n_e_file,n_e_loss);
 printf(" done\n");

 if((!strcmp(l_file,""))||(c_step))
 {

  /* this should be obsolete - z is high enough - we check now! */
  if(c_profile[d3_l.max_z-1]!=0)
  {
   printf("max_z (%d) too small - sputtering not possible\n",d3_l.max_z);
   do_sputter=0;
  }
  /* calculate ratio of c_simwindow / c_total */
  if(!do_sputter)
  {
   get_c_ratio(&c_ratio,p_file,&my_info,&d3_l);
   printf("calculated c ratio: %f\n",c_ratio);
  }

  /* sputtering realy possible ?*/
  if(n_e_loss[d3_l.max_z-1]!=0)
   printf("warning: max_z (%d) too small - there may be amorphous volumes\n",d3_l.max_z);

#ifdef DEBUG_RAND
 i=0;
 while(1)
 {
#ifdef DEBUG_CP
  printf("%d\n",get_rand_reject(d3_l.max_z,ip_max,c_profile));
#endif
#ifdef DEBUG_NEL
  printf("%d\n",get_rand_reject(d3_l.max_z,ne_max,n_e_loss));
#endif
#ifdef DEBUG_NORM
  printf("%d\n",get_rand(d3_l.max_z));
#endif
#ifdef DEBUG_LINEAR
  printf("%d\n",get_rand_lgp(d3_l.max_z,1,0));
#endif
 if(i==10000000) return 1;
 i++;
 }
#endif

#ifdef MORE_PRINTF
  printf("starting simulation ... now! :)\n");
#endif

  i=(c_step?c_step:0);
  while((i<my_info.steps) && (quit==0) && (escape==0))
  {
#ifdef MORE_PRINTF
   if(i%refresh==0) printf("step: %d\n",i);
#endif
   for(j=0;j<my_info.cpi;j++)
   {
    x_c=get_rand(d3_l.max_x);
    y_c=get_rand(d3_l.max_y);
    // z_c=get_rand_reject(d3_l.max_z,ne_max,n_e_loss);
    z_c=get_rand(d3_l.max_z);
    nel_z=URAND_MAX*(1.0*n_e_loss[z_c]/ne_max);
    process_cell(&d3_l,x_c,y_c,z_c,&my_info,nel_z);
   }
   distrib_c(&d3_l,&my_info,i,ip_max,c_profile);
#ifdef USE_DFB_API
   if(i%refresh==0)
   {
    sprintf(xyz_txt,"x: %d  y: %d  z: %d (%d nm)",x+1,y+1,z+1,z*3);
    sprintf(status_txt,"status: %c",(*(d3_l.status+x+y*d3_l.max_x+z*d3_l.max_x*d3_l.max_y)&AMORPH)?'a':'c');
    sprintf(conc_txt,"carbon: %d",*(d3_l.extra+x+y*d3_l.max_x+z*d3_l.max_x*d3_l.max_y));
    sprintf(steps_txt,"step: %d",i);
    sprintf(cc_txt,"total c: %d",my_info.cc);
    get_amorphous_layer_info(&d3_l,&sai,&sacl,&eacl);
    sprintf(start_ai_txt,"start of a-inclusions: %d nm (%d)",3*sai,sai);
    sprintf(start_acl_txt,"start of a-layer: %d nm (%d)",3*sacl,sacl);
    sprintf(end_acl_txt,"end of a-layer: %d nm (%d)",3*eacl,eacl);
    d3_lattice_draw(&d3_l,x,y,z,25,arg_v,mode,0,NULL,0,NULL,0);
   }
#endif
   if(i%resave==0 && strcmp(s_file,"") && resave!=0 && i!=0)
   {
    sprintf(s_file_tmp,"%s_%d_of_%d",s_file,i,my_info.steps);
    save_to_file(s_file_tmp,&d3_l,&my_info);
#ifdef NODFB
    printf("saved %s\n",s_file_tmp);
#endif
   }
   i++;
   if((do_sputter)&(i%my_info.s_rate==0)) sputter(&d3_l);
  }
 }

 if(strcmp(s_file,""))
 {
   printf("saved %s\n",s_file);
   save_to_file(s_file,&d3_l,&my_info);
 }

#ifdef USE_DFB_API
 /* allocating buffer for pressure values */
 printf("allocating buffer for preassure values ...");
 if((d3_l.v_ptr=malloc(d3_l.max_x*d3_l.max_y*d3_l.max_z*sizeof(unsigned int)))==NULL)
 {
  puts("cannot allocate buffer for pressure values");
  return -1;
 }
 printf(" done\n");
 /* calc values */
 printf("calculating preassure values ...");
 calc_pressure(&d3_l,my_info.range);
 printf(" done\n");
 printf("finding maximum carbon concentration ...");
 max_extra=calc_max_extra(&d3_l);
 printf(" done\n");

 while((quit==0) && (escape==0) && (nowait==0))
 {
  /* bahh! */
  if(switchmode==0) mode=0;
  if(switchmode==1) mode=1;
  if(switchmode==2) mode=2;
  if(switchmode==3) mode=3;
  /* end of bahh! */
  sprintf(xyz_txt,"x: %d  y: %d  z: %d (%d nm)",x+1,y+1,z+1,z*3);
  sprintf(status_txt,"status: %c",(*(d3_l.status+x+y*d3_l.max_x+z*d3_l.max_x*d3_l.max_y)&AMORPH)?'a':'c');
  sprintf(conc_txt,"carbon: %d",*(d3_l.extra+x+y*d3_l.max_x+z*d3_l.max_x*d3_l.max_y));
  sprintf(steps_txt,"step: end");
  sprintf(cc_txt,"total c: %d",my_info.cc);
  if(switchmode==0) strcpy(mode_txt,"view: a/c mode");
  if(switchmode==1) strcpy(mode_txt,"view: c conc mode");
  if(switchmode==2) strcpy(mode_txt,"view: a pressure mode");
  if(switchmode==3) strcpy(mode_txt,"view: a/c + profiles mode");
  d3_lattice_draw(&d3_l,x,y,z,25,arg_v,mode,max_extra,c_profile,ip_max,n_e_loss,ne_max);
  bmp=0;
  ac_distr=0;
  scan_event(&d3_l,&x,&y,&z,&quit,&escape,&switchmode,&bmp,&ac_distr);
  if(bmp) write_bmp(&d3_l,bmp,x,y,z,max_extra);
  if(ac_distr) write_ac_distr(&d3_l,ac_distr);
 }

 d3_lattice_release(&d3_l);
#endif

 return 1;
}
